# from .topic_utils import delete_topics, produce_input, consume_output
# from .setup_teardown_utils import setup_teardown_env_vars, setup_app, teardown_app, setup_teardown_source_test_data, setup_teardown_kafka_test_data, initialize_timestamp_topic
# from .validation_utils import actual_kafka_results, expected_kafka_results, validate_source_results, validate_kafka_results
# from .eloqua_utils import contact_field_map, create_contact_field_map, create_delete_contacts, retrieve_contacts_for_validation, retrieved_contact_field_map



import json
import logging
import os
import pty
import select
import subprocess
from multiprocessing import Process
from os import environ
from pathlib import Path
from shutil import rmtree, copy2
from time import sleep
from typing import List
from unittest.mock import MagicMock, patch

from confluent_kafka import KafkaException, TopicPartition
from confluent_kafka.admin import AdminClient, NewTopic
from confluent_kafka.error import KafkaError
from dotenv import load_dotenv
from virtualenv import cli_run
from virtualenvapi.manage import VirtualEnvironment

from nubium_utils.confluent_utils import get_producers, produce_message
from nubium_utils.confluent_utils.confluent_configs import init_schema_registry_configs
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import NoMessageError
from nubium_utils.yaml_parser import load_yaml_fp
from .transaction_utils import GtfoBatchApp

LOGGER = logging.getLogger(__name__)


class DudeFinished(Exception):
    pass


class DudeConsumer(GtfoBatchApp):
    def __init__(self, app_function, consume_topics_list, *args, cluster_configs=None, topic_configs=None, **kwargs):
        if cluster_configs:
            environ['NU_KAFKA_CLUSTERS_CONFIGS_YAML'] = json.dumps(cluster_configs)
            env_vars._reload()
        if topic_configs:
            environ['NU_TOPIC_CONFIGS_YAML'] = json.dumps(topic_configs)
            env_vars._reload()
        self.all_messages = []  # allows you to do batches still, and return transformed versions of the messages
        super().__init__(app_function, consume_topics_list, metrics_manager=MagicMock(), *args, **kwargs)

    def seek_consumer_to_beginning(self):
        LOGGER.info('Setting up consumer to pull from the beginning of the topics...')
        self.consumer.poll(5)
        partitions = self.consumer.assignment()
        LOGGER.debug('Seeking all partitions to 0...')
        for partition in partitions:
            offset = self.consumer.get_watermark_offsets(partition)[0]
            topic = partition.topic
            p = partition.partition
            LOGGER.debug(f'Seeking {topic} p{p} to offset {offset}')
            self.consumer.seek(TopicPartition(topic=topic, partition=p, offset=offset))

    def _app_run_loop(self, *args, **kwargs):
        messages = []
        try:
            self.consume(*args, timeout=10, **kwargs)
            if self.app_function:
                messages = self.app_function(self.transaction, *self.app_function_arglist) # allows you to transform msgs to desired result if you'd like
            if not messages:  # if app_function does not return anything (maybe they passed a placeholder function or something)
                messages = self.transaction.messages()
            self.all_messages += messages
            if not self.transaction._committed:
                self.transaction.commit()
            if len(messages) < self.transaction._default_consume_max_count:
                self.producer.poll(0)
                raise DudeFinished
        except NoMessageError:
            self.producer.poll(0)
            LOGGER.debug('All messages pulled!')
            raise DudeFinished
        except Exception as e:
            LOGGER.error(e)
            raise

    def run(self):
        self.seek_consumer_to_beginning()
        LOGGER.debug('Beginning consumption loop...')
        try:
            super().run()
        except DudeFinished:
            return self.all_messages


class KafkaToolbox:
    """
    Helpful functions for interacting with Kafka (mainly RHOSAK).
    Allows you to interface with multiple clusters at once by default assuming they are defined via your environment
    variables NUBIUM_CLUSTER_{N}. Actions include:

    Topics (multiple): create, delete, list

    Messages (multiple): produce, consume

    App: run, sync, build reqs, unit tests, integration tests
    """

    def __init__(self, cluster_configs=None, topic_configs=None, auto_configure=True):
        self.admin_clients = {}
        self.producers = {}
        self.consumer = {}
        self.schema_registry = None

        if cluster_configs:
            environ['NU_KAFKA_CLUSTERS_CONFIGS_YAML'] = json.dumps(cluster_configs)
            env_vars._reload()
        else:
            cluster_configs = load_yaml_fp(env_vars()['NU_KAFKA_CLUSTERS_CONFIGS_YAML'])
        self.cluster_configs = cluster_configs
        if topic_configs is not None:
            environ['NU_TOPIC_CONFIGS_YAML'] = json.dumps(topic_configs)
            env_vars._reload()
        else:
            topic_configs = load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])
        self.topic_configs = topic_configs
        if self.cluster_configs and auto_configure:
            self.add_admin_cluster_clients(self.cluster_configs)

    def add_admin_cluster_clients(self, configs):
        """
        Add a new client connection for a given cluster.
        Added to 'self.admin_clients', where k:v is cluster_bootstrap_url: cluster_client_instance.

        By default, uses your environment to establish the client aside from the cluster URLS.

        You can provide configs in the following ways:
            {cluster1_confluent_config_dict}
            'cluster1_bootstrap_url'
        or lists of either.
        """
        if isinstance(configs, str):
            if self.cluster_configs: # "config" passed was just a cluster name
                configs = {configs: self.cluster_configs[configs]}
            else:
                raise Exception(f'Cant create admin client for {configs}; no cluster config metadata was provided for that name.')
        for cluster_name, cluster_config in configs.items():
            if cluster_name not in self.admin_clients:
                sleep(.2)
                apply_config = {
                    "bootstrap.servers": cluster_config['url']
                }
                if 'localhost' not in cluster_config['url']:
                    apply_config.update({
                        "sasl.username": cluster_config['username'],
                        "sasl.password": cluster_config['password'],
                        "security.protocol": "sasl_ssl",
                        "sasl.mechanisms": "PLAIN",
                    })
                client = AdminClient(apply_config)
                self.admin_clients[cluster_name] = client

    def create_all_topics(self):
        """Creates all the topics by passing the whole topic map to create_topics."""
        self.create_topics(self.topic_configs)

    def create_topics(self, topic_cluster_dict, num_partitions=3, replication_factor=3, topic_config=None):
        """
        Accepts a "topic_cluster_dict" like: {'topic_a': 'cluster_x', 'topic_b': 'cluster_y', 'topic_c': ''},
        or list of topics.
        """
        defaults = {'partitions': num_partitions, 'replication_factor': replication_factor, 'config': {}}
        if "hack" in topic_cluster_dict:  # TODO: get rid of this (exists cuz of dude stuff)
            del topic_cluster_dict['hack']
        if isinstance(topic_cluster_dict, list):
            topic_cluster_dict = {k: self.topic_configs[k] for k in topic_cluster_dict}
        else:
            topic_cluster_dict = {topic: {'cluster': cfgs if isinstance(cfgs, str) else cfgs['cluster'], 'configs': self.topic_configs.get(topic, {'configs': {}})['configs']} for topic, cfgs in topic_cluster_dict.items()}
        topic_dict = {}
        for topic, configuration in topic_cluster_dict.items():
            configs = configuration['configs']
            configs.update({k: v for k, v in defaults.items() if k not in configs})
            topic_dict[configuration['cluster']] = topic_dict.get(configuration['cluster'], []) + [
                NewTopic(
                    topic=topic,
                    num_partitions=configs['partitions'],
                    replication_factor=configs['replication_factor'],
                    config=configs['config'])]
        for cluster, topic_creates in topic_dict.items():
            self.add_admin_cluster_clients(cluster)
            for topic in topic_creates:
                try:
                    _wait_on_futures_map(self.admin_clients[cluster].create_topics([topic]))
                    print(f"Topic created: {topic}")
                except KafkaException as e:
                    if e.args[0].code() == KafkaError.TOPIC_ALREADY_EXISTS:
                        print(f"Topic already exists: {topic}")
                        pass
                    else:
                        raise

    def delete_topics(self, topic_cluster_dict):
        if isinstance(topic_cluster_dict, list):
            topic_cluster_list = topic_cluster_dict
            topic_cluster_dict = {}
            all_topics = {}
            for cluster, client in self.admin_clients.items():
                all_topics.update({topic: cluster for topic in client.list_topics().topics.keys() if topic in topic_cluster_list})
            for topic in topic_cluster_list:
                topic_cluster_dict[topic] = self.topic_configs[topic]["cluster"] if topic in self.topic_configs else all_topics[topic]
        for topic, cluster in topic_cluster_dict.items():
            self.add_admin_cluster_clients(cluster)
            try:
                _wait_on_futures_map(self.admin_clients[cluster].delete_topics([topic]))
                print(f"Topic deleted: {topic}")
            except KafkaException as e:
                if e.args[0].code() == KafkaError.UNKNOWN_TOPIC_OR_PART:
                    print(f"Topic deletion failed (likely because it didn't exist): {topic}")
                    pass

    def list_topics(self, by_topic=False, mirrors=False, cluster_name=None):
        """
        Allows you to list all topics across all cluster instances.

        If you want to return the output in a format that is used by the other toolbox functions, then
        you can change "by_topic"=True; the default format makes it easier to read.

        Will not include the mirrored topics by default, but can toggle with "mirrors".
        """
        def _mirror_include(topic):
            if topic.startswith('nubium_'):
                return mirrors
            return True

        def _valid(topic):
            return not topic.startswith('__') and 'schema' not in topic

        if cluster_name:
            self.add_admin_cluster_clients(cluster_name)
        else:
            self.add_admin_cluster_clients(self.cluster_configs)
        by_cluster = {clust: [topic for topic in list(client.list_topics().topics) if _valid(topic) and _mirror_include(topic)]
                      for clust, client in self.admin_clients.items()}
        if by_topic:
            topics = {topic: clust for clust, topics in by_cluster.items() for topic in topics}
            return {k: topics[k] for k in sorted(topics.keys())}
        return {k: sorted(by_cluster[k]) for k in sorted(by_cluster.keys())}

    def produce_messages(self, topic, message_list, schema=None, cluster_name=None):
        """
        Produce a list of messages (each of which is a dict with entries 'key', 'value', 'headers') to a topic.

        Must provide the schema if you haven't produced to that topic on this instance, or if using dude CLI.
        """
        # TODO extra producer error handling?
        if not self.producers.get(topic):
            if not schema:
                raise ValueError('Schema not provided and the topic producer instance was not previously configured. '
                                 'Please provide the schema!')
            if not cluster_name:
                cluster_name = self.topic_configs[topic]['cluster']
            self.add_admin_cluster_clients(cluster_name)
            if not self.schema_registry:
                self.schema_registry = init_schema_registry_configs(as_registry_object=True)
                self.producers.update(get_producers({topic: schema}, cluster_name, self.schema_registry, return_as_singular=False))
        producer = self.producers[topic]
        for message in message_list:
            produce_message(
                producer=producer,
                producer_kwargs=message,
                metrics_manager=MagicMock())  # TODO: change NU to allow a None here? dunno
        producer.flush()

    def consume_messages(self, topics, transform_function=None, consumer=None):
        """
        Consume a list of messages (each of which is a dict with entries 'key', 'value', 'headers') from a topic.
        """
        # TODO: add easy way of changing the consumer group on the fly
        if isinstance(topics, str):
            topics = topics.split(',')
        dude_app = DudeConsumer(
            app_function=transform_function,
            consume_topics_list=topics,
            consumer=consumer,
            produce_topic_schema_dict={topic: "string" for topic in topics} # throwaway until I convert it to accept empty dict for consuming only.
        )
        return dude_app.run()

    def run_app(self, skip_topic_creation=None, skip_sync=None, run_args="", runtime_env_overrides=None):
        if _is_nubium_app():
            if not skip_sync:
                _sync_virtual_environment()
            # TODO load DUDE_APP_VENV from dude's ConfigManager
            venv_path = Path(os.environ.get("DUDE_APP_VENV", "./venv"))
            load_dotenv(Path(f'{os.path.abspath(venv_path)}/.env'), override=True)
            if not skip_topic_creation:
                topics = {os.environ[var]: "" for var in os.environ if
                          var.endswith("_TOPIC") or var.endswith("_TOPICS")}
                try:
                    LOGGER.info("Attempting to create topics...")
                    cluster = next(iter(self.admin_clients))
                    self.create_topics({topic: cluster for topic in topics})
                    sleep(5)
                except Exception as e:
                    LOGGER.info(f"Topic creation failed (likely because they already exist): {e}")
                    pass
            if not [f for f in run_args if '.py' in f]:
                run_args = ['app.py'] + list(run_args)
            if _is_faust_app():
                run_args += ["worker", "-l", "info"]
            process = Process(target=run_command_in_virtual_environment, args=("python3.8", run_args))
            if not runtime_env_overrides:
                runtime_env_overrides = {}
            with patch.dict('os.environ', runtime_env_overrides):
                process.start()
            return process.pid

    @staticmethod
    def sync_venv(wipe_existing):
        if _is_nubium_app():
            _sync_virtual_environment(wipe_existing=wipe_existing)

    @staticmethod
    def build_reqs():
        if _is_nubium_app() and _has_requirements_in():
            _ensure_virtual_environment_exists()
            run_command_in_virtual_environment("pip-compile")

    @staticmethod
    def run_unit_tests():
        if _is_nubium_app():
            _sync_virtual_environment()
            run_command_in_virtual_environment("pytest", ["./tests/unit/", "-rA", "-v"])

    @staticmethod
    def run_integration_tests():
        if _is_nubium_app():
            _sync_virtual_environment()
            venv_path = Path(os.environ.get("DUDE_APP_VENV", "./venv"))
            load_dotenv(Path(f'{os.path.abspath(venv_path)}/.env'), override=True)
            run_command_in_virtual_environment("pytest", ["./tests/integration/", "-rA", "-svv", "--log-cli-level=INFO"])


def _wait_on_futures_map(futures):
    for future in futures.values():
        future.result()
        assert future.done()
        sleep(.1)


def _is_nubium_app():
    return Path("app.py").is_file()


def _is_faust_app():
    return Path("kiwf_openshift_health_check.py").is_file()


def _has_requirements_in():
    return Path("requirements.in").is_file()


def _sync_virtual_environment(wipe_existing=False):
    venv_path = _ensure_virtual_environment_exists(wipe_existing)
    run_command_in_virtual_environment("pip-sync")
    # TODO merge these .env files to support bringing in new variables without overwriting existing customizations (see dotenv's CLI list command)
    local_dotenv = Path("configs/local.env")
    venv_dotenv = Path(f"{venv_path}/.env")
    if local_dotenv.is_file() and not venv_dotenv.is_file():
        copy2(local_dotenv, venv_dotenv)


def _ensure_virtual_environment_exists(wipe_existing: bool = False) -> Path:
    venv_path = Path(os.environ.get("DUDE_APP_VENV", "./venv"))
    if wipe_existing:
        rmtree(venv_path)
    venv_path.mkdir(parents=True, exist_ok=True)
    # TODO load python version from configuration
    cli_run([str(venv_path), "-p", "python3.8"])
    venv = VirtualEnvironment(str(venv_path), readonly=True)
    if not venv.is_installed("pip-tools"):
        # TODO lock versions of pip-tools and dependencies
        run_command_in_virtual_environment("pip", args=["install", "pip-tools"])
    return venv_path


def run_command_in_virtual_environment(command: str = "", args: List[str] = None):
    if not args:
        args = []
    venv_path = Path(os.environ.get("DUDE_APP_VENV", "./venv"))
    run_command_in_pseudo_tty(command=str(venv_path / "bin" / command), args=args)


def run_command_in_pseudo_tty(
        command: str,
        args: List[str] = None,
        output_handler=lambda data: print(data.decode("utf-8"), end=""),
        buffer_limit=512,
        buffer_timeout_seconds=0.04,
):
    # In order to get colored output from a command, the process is unfortunately quite involved.
    # Constructing a pseudo terminal interface is necessary to fake most commands into thinking they are in an environment that supports colored output
    if not args:
        args = []

    # It's possible to handle stderr separately by adding p.stderr.fileno() to the rlist in select(), and setting stderr=subproccess.PIPE
    # which would enable directing stderr to click.echo(err=True)
    # probably not worth the additional headache
    master_fd, slave_fd = pty.openpty()
    proc = subprocess.Popen([command] + args, stdin=slave_fd, stdout=slave_fd, stderr=subprocess.STDOUT, close_fds=True)

    def is_proc_still_alive():
        return proc.poll() is not None

    while True:
        ready, _, _ = select.select([master_fd], [], [], buffer_timeout_seconds)
        if ready:
            data = os.read(master_fd, buffer_limit)
            output_handler(data)
        elif is_proc_still_alive():  # select timeout
            assert not select.select([master_fd], [], [], 0)[0]  # detect race condition
            break  # proc exited
    os.close(slave_fd)  # can't do it sooner: it leads to errno.EIO error
    os.close(master_fd)
    proc.wait()



import json

from confluent_kafka import TopicPartition, KafkaException, DeserializingConsumer
from nubium_utils.general_utils import parse_headers, log_and_raise_error
from nubium_utils.custom_exceptions import NoMessageError, SignalRaise, RetryTopicSend, FailureTopicSend, MaxRetriesReached
from nubium_utils.metrics import MetricsManager
from nubium_utils.yaml_parser import load_yaml_fp
from .consumer_utils import consume_message
from .producer_utils import produce_message, get_producers
from .message_utils import shutdown_cleanup
from .rocksdb_utils import RDB
from .confluent_runtime_vars import env_vars
from .confluent_configs import init_transactional_consumer_configs, init_schema_registry_configs, get_kafka_configs, init_metrics_pushing
import logging
from json import dumps, loads
from copy import deepcopy
from os import environ, remove
from datetime import datetime
# TODO: consider using orjson for changelog speed increases?
# from orjson import dumps, loads
# changelog_schema = {"type": "bytes"}
changelog_schema = {"type": "string"}

LOGGER = logging.getLogger(__name__)


class RunTableRecovery(Exception):
    def __init__(self):
        pass


class Transaction:
    def __init__(self, producer, consumer, metrics_manager=None, auto_consume=True, timeout=None, message=None):
        self.producer = producer
        self.consumer = consumer
        self.message = message
        self.metrics_manager = metrics_manager
        if not timeout:
            timeout = 5
        self._timeout = timeout
        self.auto_consume(self.message, auto_consume)

        # mostly used for signaling to GtfoApp that there's currently a transaction and handling automatic committing
        self._committed = False
        self._active_transaction = False

    def auto_consume(self, has_input, should_consume):
        """ Consume message if initialized without one (and allowed to) """
        if not has_input and should_consume:
            self.consume()

    def messages(self):  # here for a consistent way to access message(s) currently being managed for when you subclass
        """ For a standardized way to access message(s) consumed pertaining to this transaction """
        return self.message

    def consume(self):
        self.message = consume_message(self.consumer, self.metrics_manager, self._timeout)

    def key(self):
        return self.message.key()

    def value(self):
        return self.message.value()

    def headers(self):
        return parse_headers(self.message.headers())

    def topic(self):
        return self.message.topic()

    def partition(self):
        return self.message.partition()

    def offset(self):
        return self.message.offset()
    
    def _init_transaction(self):
        """ Mark that a transaction is now underway. Triggered by trying to produce or commit a message """
        if not self._active_transaction and not self._committed:
            self.producer.begin_transaction()
            self._active_transaction = True

    def produce(self, producer_kwargs):
        self.producer.poll(0)
        self._init_transaction()
        produce_message(self.producer, producer_kwargs, self.metrics_manager, self.headers())
        self.producer.poll(0)

    def produce_retry(self, exception=None):
        retry_topic = None
        headers = self.headers()
        guid = headers['guid']
        kafka_retry_count = int(headers.get('kafka_retry_count', '0'))

        if kafka_retry_count < int(env_vars()['NU_RETRY_COUNT_MAX']):
            headers['kafka_retry_count'] = str(kafka_retry_count + 1)
            retry_topic = env_vars()['NU_CONSUME_TOPICS']
        else:
            headers['kafka_retry_count'] = '0'
            retry_topic = env_vars().get('NU_PRODUCE_RETRY_TOPICS', '')

        if retry_topic:
            if not exception:
                exception = RetryTopicSend()
            LOGGER.warning('; '.join([str(exception), f'retrying GUID {guid}']))
            self.produce(dict(
                topic=retry_topic,
                value=self.value(),
                key=self.key(),
                headers=headers))
        else:
            if not exception:
                exception = FailureTopicSend()
            LOGGER.error('; '.join([str(exception), f'GUID {guid}']))
            self.produce_failure(exception=MaxRetriesReached())

    def produce_failure(self, exception=None):
        headers = self.headers()
        guid = headers['guid']
        headers['kafka_retry_count'] = '0'
        failure_topic = env_vars()['NU_PRODUCE_FAILURE_TOPICS']

        if not exception:
            exception = FailureTopicSend()
        LOGGER.error('; '.join([type(exception).__name__, str(exception), f'failing GUID {guid}']))
        headers["exception"] = json.dumps({"name": type(exception).__name__, "description": str(exception)})

        LOGGER.debug(f'Adding a message to the produce queue for deadletter/failure topic {env_vars()["NU_PRODUCE_FAILURE_TOPICS"]}')
        self.produce(dict(
            topic=failure_topic,
            value=self.value(),
            key=self.key(),
            headers=headers))
        LOGGER.info(f'Message added to the deadletter/failure topic produce queue; GUID {guid}')

    def commit(self, mark_committed=True):
        """ Allows manual commits (safety measures in place so that you cant commit the same message twice)."""
        self._init_transaction()
        if self._active_transaction:
            if self.message:
                self.producer.send_offsets_to_transaction(
                    [TopicPartition(self.topic(), self.partition(), self.offset() + 1)], self.consumer.consumer_group_metadata())
            self.producer.commit_transaction()
            self.producer.poll(0)
            self._committed = mark_committed
            if self._committed:
                self._active_transaction = False
                LOGGER.debug('Transaction Committed!')


class GtfoApp:
    """ The main class to use for most GTFO apps. See README for initialization/usage details. """
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict=None, transaction_type=Transaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None):
        self.transaction_type = transaction_type
        self.transaction = None
        self.app_function = app_function
        if not app_function_arglist:
            app_function_arglist = []
        self.app_function_arglist = app_function_arglist
        if not metrics_manager:
            metrics_manager = MetricsManager()
        self.metrics_manager = metrics_manager
        init_metrics_pushing(self.metrics_manager)
        if isinstance(consume_topics_list, str):
            consume_topics_list = consume_topics_list.split(',')
        if not produce_topic_schema_dict:  # for when the app is consume-only
            produce_topic_schema_dict = {topic: None for topic in consume_topics_list}
        if not schema_registry:
            schema_registry = init_schema_registry_configs(as_registry_object=True)
        if not cluster_name:
            cluster_name = self._get_cluster_name(consume_topics_list)
        if not consumer:
            consumer = self._get_transactional_consumer(consume_topics_list, schema_registry, cluster_name)
        self.consumer = consumer
        if not producer:
            producer = self._get_transactional_producer(produce_topic_schema_dict, schema_registry, cluster_name)
        self.producer = producer

    def _get_cluster_name(self, consume_topics_list):
        topic = consume_topics_list[0] if isinstance(consume_topics_list, list) else consume_topics_list
        return load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])[topic]['cluster']

    def _get_transactional_producer(self, topic_schema_dict, schema_registry, cluster_name):
        LOGGER.debug('Setting up Kafka Transactional Producer')
        producer = get_producers(topic_schema_dict, cluster_name, schema_registry, transactional=True)
        producer.init_transactions()
        LOGGER.debug('Producer setup complete.')
        return producer

    def _get_transactional_consumer(self, topics, schema_registry, cluster_name, default_schema=None, auto_subscribe=True):
        LOGGER.debug('Setting up Kafka Transactional Consumer')
        consumer = DeserializingConsumer(
            init_transactional_consumer_configs(topics, schema_registry, get_kafka_configs(cluster_name)[0], cluster_name, default_schema))
        if auto_subscribe:
            consumer.subscribe(topics)  # in case multiple topics are read from
            LOGGER.info(f'Transactional consumer subscribed to topics:\n{topics}')
        return consumer

    def consume(self, *args, timeout=10, **kwargs):
        """
        Accepts *args and **kwargs to make it easy to alter functionality of this function in subclasses.
        Public method so that you can manually consume messages if you desire; helpful for debugging.
        """
        LOGGER.debug('App - Attempting to consume')
        self.transaction = self.transaction_type(self.producer, self.consumer, *args, metrics_manager=self.metrics_manager, timeout=timeout, **kwargs)
        return self.transaction

    def _app_run_loop(self, *args, **kwargs):
        try:
            self.consume(*args, **kwargs)
            self.app_function(self.transaction, *self.app_function_arglist)
            if not self.transaction._committed:
                self.transaction.commit()
        except NoMessageError:
            self.producer.poll(0)
            LOGGER.debug('No messages!')

    def kafka_cleanup(self):
        """ Public method in the rare cases where you need to do some cleanup on the consumer object manually. """
        shutdown_cleanup(consumer=self.consumer)

    def _app_shutdown(self):
        LOGGER.info('App is shutting down...')
        try:
            if self.transaction._active_transaction:
                self.producer.abort_transaction(10)
        except:
            pass
        finally:
            self.kafka_cleanup()

    def run(self, *args, health_path='/tmp', **kwargs):
        with open(f'{health_path}/health', 'w') as health_file:
            health_file.write('Healthy')
        try:
            while True:
                self._app_run_loop(*args, **kwargs)
        except SignalRaise:
            LOGGER.info('Shutdown requested!')
        except Exception as e:
            LOGGER.error(e)
            if self.metrics_manager:
                log_and_raise_error(self.metrics_manager, e)
        finally:
            self._app_shutdown()
            remove(f'{health_path}/health')


class BatchTransaction(Transaction):
    def __init__(self, producer, consumer, metrics_manager=None, auto_consume=True, timeout=None, message_batch=None):
        self._default_consume_max_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT'])
        if not message_batch:
            message_batch = []
        self.message_batch = message_batch
        self.message_batch_partition_offset_msg = {}
        message = self.message_batch[-1] if self.message_batch else None
        super().__init__(producer, consumer, metrics_manager=metrics_manager, auto_consume=False, timeout=timeout, message=message)
        self.auto_consume(self.message_batch, auto_consume)

    def messages(self):
        return self.message_batch

    def consume(self, time_elapse_max_seconds=None, consume_max_count=None):
        """
        Allows you to (additionally) consume more messages on demand via the consume_max_count.
        Usually only necessary with more complex consumption patterns where bulk is swapped between.

        Note the "max count" is the max size you allow self.message_batch to be, which includes all previous consumes
        on this transaction.
        """
        if time_elapse_max_seconds:
            time_elapse_start = datetime.now().timestamp()
        if not consume_max_count:
            consume_max_count = self._default_consume_max_count

        def max_time_elapsed():
            if time_elapse_max_seconds:
                return datetime.now().timestamp() - time_elapse_start > time_elapse_max_seconds
            return False
            
        def keep_consuming():
            if not consume_max_count or max_time_elapsed():
                return True
            return len(self.message_batch) < consume_max_count

        LOGGER.debug(f'Consuming up to {consume_max_count} messages!')
        while keep_consuming():
            try:
                super().consume()
                self.message_batch.append(self.message)
                self.message_batch_partition_offset_msg[self.message.partition()] = self.message
            except NoMessageError:
                if not self.message_batch:
                    raise
                else:
                    break
        LOGGER.debug(f'Finished batch consumption; total messages: {len(self.message_batch)}')

    def produce_retry(self, exception=None, message=None):
        if message:
            self.message = message
            super().produce_retry(exception=exception)
        else:
            for msg in self.message_batch:
                self.message = msg
                super().produce_retry(exception=exception)

    def produce_failure(self, exception=None, message=None):
        if message:
            self.message = message
            super().produce_failure(exception=exception)
        else:
            for msg in self.message_batch:
                self.message = msg
                super().produce_failure(exception=exception)

    def commit(self, mark_committed=True):
        for msg in self.message_batch_partition_offset_msg.values():
            self.message = msg
            LOGGER.debug(f'finalizing transaction for messages on partition {self.partition()}')
            super().commit(mark_committed=False)
        if mark_committed:
            self._committed = True
        if self._committed:
            self._active_transaction = False
        LOGGER.debug('All partition transactions committed!')


class GtfoBatchApp(GtfoApp):
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict=None, transaction_type=BatchTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None):
        super().__init__(app_function, consume_topics_list, produce_topic_schema_dict=produce_topic_schema_dict, transaction_type=transaction_type,
                         app_function_arglist=app_function_arglist, metrics_manager=metrics_manager, schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer, producer=producer)


class TableTransaction(Transaction):
    def __init__(self, producer, consumer, app_changelog_topic, app_rdb_tables, metrics_manager=None, message=None, auto_consume=True, timeout=5):
        self.app_changelog_topic = app_changelog_topic
        self.app_rdb_tables = app_rdb_tables
        self._changelog_updated = False
        self._pending_table_write = None
        super().__init__(producer, consumer, metrics_manager=metrics_manager, message=message, auto_consume=auto_consume, timeout=timeout)

    def read_table_entry(self):
        return self._rdb_read()

    def update_table_entry(self, value):
        self._pending_table_write = deepcopy(value)
        if isinstance(self._pending_table_write, (list, dict)):
            # LOGGER.debug(f'attempting json dumps of {self._pending_table_write}')
            self._pending_table_write = dumps(self._pending_table_write)

    def delete_table_entry(self):
        self._pending_table_write = '-DELETED-'

    def _update_changelog(self):
        self.produce(dict(
            topic=self.app_changelog_topic,
            key=self.key(),
            value=self._pending_table_write
        ))
        self._changelog_updated = True
        self.producer.poll(0)

    def _recover_table_via_changelog(self):
        value = self.value()
        try:
            value = loads(value)
        except:
            pass
        if value == '-DELETED-':
            self.delete_table_entry()
        else:
            self.update_table_entry(value)
        super().commit()
        self._rdb_write()

    def _rdb_write(self):
        if self._pending_table_write == '-DELETED-':
            LOGGER.debug('Finalizing table entry delete...')
            self.app_rdb_tables[self.partition()].delete(self.key())
            self.app_rdb_tables[self.partition()].write('offset', str(self._rdb_offset() + 2))
        else:
            LOGGER.debug(f'Finalizing table entry write:\npartition{self.partition()},\nkey:{self.key()}')
            self.app_rdb_tables[self.partition()].write_batch(
                {self.key(): self._pending_table_write,
                 'offset': str(self._rdb_offset() + 2)})

    def _rdb_read(self):
        value = self.app_rdb_tables[self.partition()].read(self.key())
        try:
            value = loads(value)
        except:
            pass
        LOGGER.debug(f'Read table value: {value}')
        return value

    def _rdb_offset(self):
        value = self.app_rdb_tables[self.partition()].read('offset')
        if not value:
            value = self.offset() if self.offset() else 0
        return int(value)

    def commit(self, mark_committed=True):
        if not self._changelog_updated and self._pending_table_write:
            self._update_changelog()
        super().commit(mark_committed=False)
        if self._pending_table_write:
            self._rdb_write()
        if mark_committed:
            self._committed = True
            self._active_transaction = False
        self._pending_table_write = None
        LOGGER.debug('Transaction Committed!')


class GtfoTableApp(GtfoApp):
    def __init__(self, app_function, consume_topic, produce_topic_schema_dict=None, transaction_type=TableTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None):
        self.changelog_topic = f"{environ['NU_APP_NAME']}__changelog"
        if not produce_topic_schema_dict:
            produce_topic_schema_dict = {}
        if self.changelog_topic not in produce_topic_schema_dict:
            produce_topic_schema_dict.update({self.changelog_topic: changelog_schema})
        self.consume_topic = consume_topic
        self.rdb_tables = {}
        self._pending_primary_partitions = {}
        self._pending_table_recoveries = {}
        if not cluster_name:
            cluster_name = self._get_cluster_name(self.consume_topic)
        if not schema_registry:
            schema_registry = init_schema_registry_configs(as_registry_object=True)
        if not consumer:
            consumer = self._set_table_consumer(self.consume_topic, schema_registry, cluster_name=cluster_name)
        super().__init__(
            app_function, self.consume_topic, produce_topic_schema_dict, transaction_type=transaction_type,
            app_function_arglist=app_function_arglist, metrics_manager=metrics_manager, schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer, producer=producer)

    def _rdb_close(self, partitions=None):
        full_shutdown = False
        if not partitions:
            partitions = list(self.rdb_tables.keys())
            full_shutdown = True
        LOGGER.debug(f'RocksDB - closing connections for partitions {partitions}')
        for p in partitions:
            try:
                self.rdb_tables[p].close()
                del self.rdb_tables[p]
                LOGGER.debug(f'p{p} RDB table connection closed.')
            except KeyError:
                if not full_shutdown:
                    LOGGER.debug(
                        f'RDB Table p{p} did not seem to be mounted and thus could not unmount,'
                        f' likely caused by multiple rebalances in quick succession.'
                        f' This is unliklely to cause issues as the client is in the middle of adjusting itself, '
                        f' but should be noted.')
        LOGGER.info(f'RocksDB - closed connections for partitions {partitions}')

    def _confirm_proper_assignment(self):
        assignments = self.consumer.assignment()
        if self.changelog_topic in [p_obj.topic for p_obj in assignments]:
            raise Exception('changelog topic was included in normal consumption assignment! ABANDON SHIP!!!')
        topic = set([p_obj.partition for p_obj in assignments])
        tables = set(self.rdb_tables)
        if topic != tables:
            LOGGER.warning(f"Partition/Table assignment mismatch!! topic: {topic}; tables: {tables}")

    def _resume_consumption(self, from_recovery=False):
        if from_recovery:
            changelog_unassign = [p_obj for p_obj in self.consumer.assignment() if p_obj.topic == self.changelog_topic]
            LOGGER.debug(f'unassigning changelog partitions: {changelog_unassign}')
            self.consumer.incremental_unassign(changelog_unassign)
            LOGGER.debug(f'Resuming consumption for partitions:\n{list(self._pending_primary_partitions.values())}')
            self.consumer.resume(list(self._pending_primary_partitions.values()))
        self._confirm_proper_assignment()
        self._pending_table_recoveries = {}  # for tracking all recovery partitions during rebalances (based on primary)
        self._pending_primary_partitions = {}  # for tracking all assigned partitions during rebalances
        LOGGER.info('Continuing normal consumption loop...')

    def _rdb_init(self, partition):
        self.rdb_tables[partition] = RDB(f'p{partition}')
        LOGGER.debug(f'RDB table for p{partition} initialized')

    def _get_changelog_watermarks(self, recovery_partitions):
        """
        Note: this is a separate function since it requires the consumer to communicate with the broker
        """
        LOGGER.debug(f'Recovery_items: {recovery_partitions.items()}')
        return {p: self.consumer.get_watermark_offsets(p_obj) for p, p_obj in recovery_partitions.items()}

    def _refresh_pending_table_recoveries(self):
        """
        confirms new recoveries and removes old ones if not applicable anymore
        """
        recovery_partitions = {p: TopicPartition(topic=self.changelog_topic, partition=p) for p, p_obj in self._pending_primary_partitions.items()}
        tables_to_recover = {}
        for partition, watermarks in self._get_changelog_watermarks(recovery_partitions).items():
            LOGGER.debug(f'(lowwater, highwater) for changelog p{partition}: {watermarks}')
            if watermarks[0] != watermarks[1]:
                table_offset = self.rdb_tables[partition].read('offset')
                if table_offset:
                    table_offset = int(table_offset)
                else:
                    table_offset = 0
                if table_offset < watermarks[1]:
                    tables_to_recover[partition] = {'table_offset': table_offset, 'watermarks': watermarks, 'partition_obj': recovery_partitions[partition]}
        self._pending_table_recoveries = tables_to_recover

    def _init_rdb_tables(self, partition_objs):
        for p in partition_objs:
            partition = p.partition
            if partition not in self.rdb_tables:
                self._rdb_init(partition)

    def _new_table_recoveries(self, new_partitions):
        """ Add new recovery partitions based on additional rebalances """
        self._refresh_pending_table_recoveries()
        return {p: table_meta for p, table_meta in self._pending_table_recoveries.items() if p in new_partitions}

    def _recover_or_resume(self):
        # TODO: check to see if handling it via exception is actually necessary anymore?
        """
        Due to confluent-kafka consumers asynchronous operations, the best way to force a desired
        control flow within the general app consumer loop is to raise an exception and
        interrupt whatever the app is currently doing and start recovering.
        """
        if self._pending_table_recoveries:
            LOGGER.debug(f'Table recovery required: {self._pending_table_recoveries}')
            raise RunTableRecovery
        else:
            LOGGER.debug('Preparing to resume consumption pattern...')
            self._resume_consumption()

    def _rocksdb_assign(self, consumer, add_partition_objs):
        """
        Called every time a rebalance happens and handles table assignment and recovery flow.
        NOTE: rebalances pass relevant partitions per rebalance call which can happen multiple times, especially when
        multiple apps join at once; we have objects to track all updated partitions received during the entire rebalance.
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _rocksdb_assign will ALWAYS be called (even when no new assignments are required) after _rocksdb_unassign.
        """
        LOGGER.info('Rebalance Triggered - Assigment')
        LOGGER.debug(f'Consumer - Assigning additional partitions: {[p_obj.partition for p_obj in add_partition_objs]}')
        if add_partition_objs:
            new_partitions = {p_obj.partition: p_obj for p_obj in add_partition_objs}
            self._pending_primary_partitions.update(new_partitions)
            self._init_rdb_tables(list(self._pending_primary_partitions.values()))
            add_table_recovery = self._new_table_recoveries(new_partitions)
            self.consumer.incremental_assign(add_partition_objs)
            if add_table_recovery:
                LOGGER.debug('New partition assignment will require table recovery...')
                self.consumer.incremental_assign([v['partition_obj'] for k, v in self._pending_table_recoveries.items() if k in new_partitions])  # could do all of them, but less exception handling/cleaner this way
        else:
            has_assignments = self.consumer.assignment()
            if has_assignments:
                LOGGER.debug('No new/additional partitions assigned.')
                LOGGER.info(f'Resuming current assignment of: {[p_obj.partition for p_obj in has_assignments]}')
            else:
                LOGGER.info('Awaiting partition assignments from broker...')
        LOGGER.info('Consumer - Assignment request complete.')
        self._recover_or_resume()

    def _rocksdb_unassign(self, consumer, drop_partition_objs):
        """
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _rocksdb_assign will always be called (even when no new assignments are required) after _rocksdb_unassign.
        """
        partitions = [p_obj.partition for p_obj in drop_partition_objs]
        LOGGER.debug(f'Consumer - Unassigning topic {self.consume_topic} partitions: {partitions}')
        self.consumer.incremental_unassign(drop_partition_objs)
        if self._pending_table_recoveries:
            LOGGER.debug(f'unassigning changelog partitions: {drop_partition_objs}')
            self.consumer.incremental_unassign([p_obj for p_obj in self.consumer.assignment() if p_obj.topic == self.changelog_topic and p_obj.partition in drop_partition_objs])
        self._pending_primary_partitions = {k: v for k, v in self._pending_primary_partitions.items() if k not in partitions}
        self._pending_table_recoveries = {k: v for k, v in self._pending_table_recoveries.items() if k not in partitions}
        LOGGER.debug(f'pending_primary_partitions after unassignment: {self._pending_primary_partitions}')
        LOGGER.debug(f'table_recovery_status after unassignment: {self._pending_table_recoveries}')
        self._rdb_close(partitions)

    def _set_table_consumer(self, topic, schema_registry, default_schema=None, cluster_name=None):
        if isinstance(topic, str):
            topic = [topic]
        consumer = self._get_transactional_consumer(topic, schema_registry, cluster_name, default_schema, False)
        consumer.subscribe(topic, on_assign=self._rocksdb_assign, on_revoke=self._rocksdb_unassign, on_lost=self._rocksdb_unassign)
        LOGGER.debug('Table consumer initialized')
        return consumer

    def _table_recovery_refresh_starting_offsets(self):
        """ Refresh the offsets of recovery partitions to account for updated recovery states during rebalancing """
        for p, offsets in self._pending_table_recoveries.items():
            new_offset = self._pending_table_recoveries[p]['table_offset']
            low_mark = self._pending_table_recoveries[p]['watermarks'][0]
            if low_mark > new_offset:  # handles offsets that have been removed/compacted. Should never happen, but ya know
                LOGGER.info(
                    f'p{p} table has an offset ({new_offset}) less than the changelog lowwater ({low_mark}), likely due to retention settings. Setting {low_mark} as offset start point.')
                new_offset = low_mark
            high_mark = self._pending_table_recoveries[p]['watermarks'][1]
            LOGGER.debug(f'p{p} table has an offset delta of {high_mark - new_offset}')
            self._pending_table_recoveries[p]['partition_obj'].offset = new_offset

    def _table_recovery_loop(self, checks=3):
        LOGGER.info('BEGINNING TABLE RECOVERY PROCEDURE')
        while checks and self._pending_table_recoveries:
            try:
                self.consume()
                LOGGER.debug(f'Recovery write is {self.transaction.value()}')
                self.transaction._recover_table_via_changelog()
                p = self.transaction.partition()
                LOGGER.debug(
                    f"transaction_offset - {self.transaction.offset() + 2}, watermark - {self._pending_table_recoveries[p]['watermarks'][1]}")
                if self._pending_table_recoveries[p]['watermarks'][1] - (self.transaction.offset() + 2) <= 0:
                    LOGGER.info(f'table partition {p} fully recovered!')
                    del self._pending_table_recoveries[p]
            except NoMessageError:
                checks -= 1
                LOGGER.debug(f'No changelog messages, checks remaining: {checks}')
        LOGGER.info("TABLE RECOVERY COMPLETE!")

    def _throwaway_poll(self):
        """ Have to poll (changelog topic) after first assignment to it to allow seeking """
        LOGGER.debug("Performing throwaway poll to allow assignments to properly initialize...")
        try:
            self.consume(timeout=3)
        except NoMessageError:
            pass

    def _table_recovery(self):
        # TODO: check if below is necc (maybe if you don't do an exception interrupt)?
        # while self._pending_table_recoveries: # In case of rebalance interruptions
        try:
            self._table_recovery_refresh_starting_offsets()
            LOGGER.debug(f'table_recovery_status before recovery attempt: {self._pending_table_recoveries}')
            for table_meta in self._pending_table_recoveries.values():
                self.consumer.seek(table_meta['partition_obj'])
            self._table_recovery_loop()
            self._resume_consumption(from_recovery=True)
        except KafkaException as e:
            if 'Failed to seek to offset' in e.args[0].str():
                LOGGER.debug('Running a consumer poll to allow seeking to work on the changelog partitions...')
                self._throwaway_poll()
                self._table_recovery()
        except Exception as e:
            LOGGER.debug(f'{[arg for arg in e.args]}')
            LOGGER.debug('Table recovery interrupted due to rebalance...retrying')

    def consume(self, *args, **kwargs):
        super().consume(*args, app_changelog_topic=self.changelog_topic, app_rdb_tables=self.rdb_tables, **kwargs)

    def _app_run_loop(self, *args, **kwargs):
        try:
            self.consume(*args, **kwargs)
            self.app_function(self.transaction, *self.app_function_arglist)
            if not self.transaction._committed:
                self.transaction.commit()
        except NoMessageError:
            self.producer.poll(0)
            LOGGER.debug('No messages!')
        except RunTableRecovery:
            self._table_recovery()

    def kafka_cleanup(self):
        super().kafka_cleanup()
        self._rdb_close()



import json
import logging
from copy import deepcopy

import pytest

LOGGER = logging.getLogger(__name__)


@pytest.fixture()
def actual_kafka_results(consume_output):
    LOGGER.info('Preparing actual kafka results...')
    return [{k: v for k, v in message.items() if k in ['key', 'value']} for message in consume_output]


@pytest.fixture()
def expected_kafka_results(map_kafka_expected_output):
    LOGGER.info('Preparing expected kafka results...')
    return [{k: v for k, v in message.items() if k in ['key', 'value']} for message in map_kafka_expected_output]


def remove_fields_from_dict(dict_to_update, fields_to_ignore):
    new_dict = deepcopy(dict_to_update)
    if hasattr(dict_to_update, 'items'):
        for key, value in dict_to_update.items():
            if key in fields_to_ignore:
                new_dict.pop(key)
            if isinstance(value, dict):
                new_value = remove_fields_from_dict(value, fields_to_ignore)
                new_dict[key] = new_value
    return new_dict


@pytest.fixture()
def validate_external_results(map_external_expected_output, map_external_actual_output, request):
    LOGGER.info('Validating external results...')
    try:
        kwargs = {'fields_to_ignore': None}
        kwargs.update(getattr(request, 'param', {}))
        if kwargs['fields_to_ignore']:
            map_external_expected_output = [remove_fields_from_dict(record, kwargs['fields_to_ignore']) for record in map_external_expected_output]
            map_external_actual_output = [remove_fields_from_dict(record, kwargs['fields_to_ignore']) for record in map_external_actual_output]
        for expected in map_external_expected_output:
            assert expected in map_external_actual_output
        LOGGER.info('TEST SUCCESS!')
    except AssertionError:
        LOGGER.info(
            f"TEST FAILURE =(\n\nEXTERNAL EXPECTED=\n{json.dumps(map_external_expected_output, indent=4)}\n\nACTUAL:\n{json.dumps(map_external_actual_output, indent=4)}")
        raise


@pytest.fixture()
def validate_kafka_results(actual_kafka_results, expected_kafka_results, request):
    LOGGER.info('Validating kafka results...')
    try:
        kwargs = {'fields_to_ignore': None}
        kwargs.update(getattr(request, 'param', {}))
        if kwargs['fields_to_ignore']:
            expected_kafka_results = [remove_fields_from_dict(record, kwargs['fields_to_ignore']) for record in expected_kafka_results]
            actual_kafka_results = [remove_fields_from_dict(record, kwargs['fields_to_ignore']) for record in actual_kafka_results]
        for expected in expected_kafka_results:
            assert expected in actual_kafka_results
        LOGGER.info('TEST SUCCESS!')
    except AssertionError:
        LOGGER.info(
            f"TEST FAILURE =(\n\nKAFKA EXPECTED=\n{json.dumps(expected_kafka_results, indent=4)}\n\nKAFKA ACTUAL=\n{json.dumps(actual_kafka_results, indent=4)}")
        raise



from .consumer_utils import consume_message
from .message_utils import handle_no_messages, shutdown_cleanup
from .producer_utils import produce_message, get_producers
from .confluent_configs import init_metrics_pushing
from .dev_toolbox import KafkaToolbox
from .transaction_utils import GtfoApp, GtfoBatchApp, GtfoTableApp


import pytest
from unittest.mock import patch, MagicMock, call
import os
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.confluent_utils.producer_utils import *
from copy import deepcopy


patch_path = 'nubium_utils.confluent_utils.producer_utils'


def required_env_vars():
    return {
        "NU_LOGLEVEL": "DEBUG",
        "NU_HOSTNAME": "test-app-0",
        "NU_APP_NAME": "test_app",
        "NU_MP_PROJECT": "none",
        "NU_MP_CLUSTER": "none",
        "NU_DO_METRICS_PUSHING": "false",
        "NU_SCHEMA_REGISTRY_URL": "sr_url",
        "NU_SCHEMA_REGISTRY_USERNAME": "sr_un",
        "NU_SCHEMA_REGISTRY_PASSWORD": "sr_pw",
        "NU_TOPIC_CONFIGS_YAML": '{"test_topic_0": {"cluster": "cluster_0"}, "test_topic_1": {"cluster": "cluster_1"}}',
        "NU_KAFKA_CLUSTERS_CONFIGS_YAML": '{"cluster_0": {"url": "www", "username": "un", "password": "pw"}}',
        "NU_CONSUMER_POLL_TIMEOUT": "5",
        "NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES": "1"
    }


@pytest.fixture(autouse=True)
def set_env_vars():
    evars = required_env_vars()
    env_patch = patch.dict('os.environ', evars)
    env_patch.start()
    env_vars._reload()


@pytest.fixture()
def topic_schema_dict():
    return {'test_topic_1': {"type": "string"}, 'test_topic_2': {"type": "string"}, 'test_topic_3': None}


@pytest.fixture()
def producer_serializer():
    return MagicMock()


@pytest.fixture()
def producer(producer_serializer):
    producer = MagicMock()
    producer.schema_dict = {'test_topic_1': producer_serializer}
    return producer


class TestProducerUtils:
    def test_producer_serializers(self, topic_schema_dict):
        schema_reg_object = MagicMock()
        with patch(f'{patch_path}.AvroSerializer') as AS:
            serializer1, serializer2 = MagicMock(), MagicMock()
            AS.side_effect = [serializer1, serializer2]
            assert producer_serializers(topic_schema_dict, schema_reg_object) == {'test_topic_1': serializer1, 'test_topic_2': serializer2}
            AS.assert_has_calls([call(schema_reg_object, '{"type": "string"}'), call(schema_reg_object, '{"type": "string"}')])

    def test_get_producers(self, topic_schema_dict):
        schema_reg_obj = MagicMock()
        cluster_name = 'test_cluster'
        mock_producer_configs = {'producer.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.SerializingProducer') as SP:
            producer1, producer2, producer3 = MagicMock(), MagicMock(), MagicMock()
            SP.side_effect = [producer1, producer2, producer3]
            with patch(f'{patch_path}.init_producer_configs') as producer_configs:
                producer_configs.return_value = mock_producer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    with patch(f'{patch_path}.producer_serializers') as serializer:
                        sdict1, sdict2, sdict3 = {'test_topic_1': MagicMock()}, {'test_topic_2': MagicMock()}, {'test_topic_2': MagicMock()}
                        serializer.side_effect = [sdict1, sdict2, sdict3]
                        expected = {'test_topic_1': producer1, 'test_topic_2': producer2, 'test_topic_3': producer3}
                        result = get_producers(topic_schema_dict, cluster_name, schema_reg_obj)
                        serializer.assert_has_calls(
                            [call({'test_topic_1': {"type": "string"}}, schema_reg_obj),
                             call({'test_topic_2': {"type": "string"}}, schema_reg_obj),
                             call({'test_topic_3': None}, schema_reg_obj)])
                    kafka_configs.assert_has_calls([
                        call(cluster_name),
                        call(cluster_name),
                        call(cluster_name)])
                producer_configs.assert_has_calls([
                    call({'test_topic_1': {"type": "string"}}, schema_reg_obj, mock_kafka_configs, cluster_name, False),
                    call({'test_topic_2': {"type": "string"}}, schema_reg_obj, mock_kafka_configs, cluster_name, False),
                    call({'test_topic_3': None}, schema_reg_obj, mock_kafka_configs, cluster_name, False)])
            SP.assert_has_calls([
                call(mock_producer_configs),
                call(mock_producer_configs),
                call(mock_producer_configs)])
        assert expected == result
        assert result['test_topic_1'] == producer1
        assert result['test_topic_2'] == producer2
        assert result['test_topic_3'] == producer3
        assert producer1.topic == 'test_topic_1'
        assert producer1.schema_dict == sdict1
        assert producer2.topic == 'test_topic_2'
        assert producer2.schema_dict == sdict2
        assert producer3.topic == 'test_topic_3'
        assert producer3.schema_dict == sdict3

    def test_get_producers__singular(self):
        topic_schema_dict = {'test_topic_1': {"type": "string"}}
        schema_reg_obj = MagicMock()
        cluster_name = 'test_cluster'
        mock_producer_configs = {'producer.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.SerializingProducer') as SP:
            producer1 = MagicMock()
            SP.side_effect = [producer1]
            with patch(f'{patch_path}.init_producer_configs') as producer_configs:
                producer_configs.return_value = mock_producer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    with patch(f'{patch_path}.producer_serializers') as serializer:
                        sdict1 = {'test_topic_1': MagicMock()}
                        serializer.side_effect = [sdict1]
                        expected = producer1
                        result = get_producers(topic_schema_dict, cluster_name, schema_reg_obj)
                        serializer.assert_has_calls([call({'test_topic_1': {"type": "string"}}, schema_reg_obj)])
                    kafka_configs.assert_has_calls([call(cluster_name)])
                producer_configs.assert_has_calls([call({'test_topic_1': {"type": "string"}}, schema_reg_obj, mock_kafka_configs, cluster_name, False)])
            SP.assert_has_calls([call(mock_producer_configs)])
        assert expected == result
        assert producer1.topic == 'test_topic_1'
        assert producer1.schema_dict == sdict1

    def test_get_producers__transactional(self):
        topic_schema_dict = {'test_topic_1': {"type": "string"}, 'test_topic_2': {"type": "string"}}
        schema_reg_obj = MagicMock()
        cluster_name = 'test_cluster'
        mock_producer_configs = {'producer.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.SerializingProducer') as SP:
            producer1, producer2, producer3 = MagicMock(), MagicMock(), MagicMock()
            SP.side_effect = [producer1]
            with patch(f'{patch_path}.init_producer_configs') as producer_configs:
                producer_configs.return_value = mock_producer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    with patch(f'{patch_path}.producer_serializers') as serializer:
                        sdict1 = {'test_topic_1': MagicMock()}
                        serializer.side_effect = [sdict1]
                        expected = producer1
                        result = get_producers(topic_schema_dict, cluster_name, schema_reg_obj, transactional=True)
                        serializer.assert_has_calls([call({'test_topic_1': {"type": "string"}, 'test_topic_2': {"type": "string"}}, schema_reg_obj)])
                    kafka_configs.assert_has_calls([call(cluster_name)])
                producer_configs.assert_has_calls([call({'test_topic_1': {"type": "string"}, 'test_topic_2': {"type": "string"}}, schema_reg_obj, mock_kafka_configs, cluster_name, True)])
            SP.assert_has_calls([call(mock_producer_configs)])
        assert expected == result
        assert producer1.schema_dict == sdict1

    def test_get_producers__as_singular_false(self):
        topic_schema_dict = {'test_topic_1': {"type": "string"}}
        schema_reg_obj = MagicMock()
        cluster_name = 'test_cluster'
        mock_producer_configs = {'producer.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.SerializingProducer') as SP:
            producer1, producer2, producer3 = MagicMock(), MagicMock(), MagicMock()
            SP.side_effect = [producer1]
            with patch(f'{patch_path}.init_producer_configs') as producer_configs:
                producer_configs.return_value = mock_producer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    with patch(f'{patch_path}.producer_serializers') as serializer:
                        sdict1 = {'test_topic_1': MagicMock()}
                        serializer.side_effect = [sdict1]
                        expected = {'test_topic_1': producer1}
                        result = get_producers(topic_schema_dict, cluster_name, schema_reg_obj, return_as_singular=False)
                        serializer.assert_has_calls([call({'test_topic_1': {"type": "string"}}, schema_reg_obj)])
                    kafka_configs.assert_has_calls([call(cluster_name)])
                producer_configs.assert_has_calls([call({'test_topic_1': {"type": "string"}}, schema_reg_obj, mock_kafka_configs, cluster_name, False)])
            SP.assert_has_calls([call(mock_producer_configs)])
        assert expected == result
        assert result['test_topic_1'] == producer1
        assert producer1.topic == 'test_topic_1'
        assert producer1.schema_dict == sdict1

    def test_produce_message__with_metric_manager(self, producer, producer_serializer):
        """
        Happy path, check that header passthrough gets the header kw arg updates
        """
        metrics_manager = MagicMock()
        test_input = [
            producer,
            dict(topic='test_topic_1',
                 key='blah_key',
                 value={'blah_value': 'blah_value'},
                 headers={'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}),
            metrics_manager,
            [('guid', b'112233'), ('last_updated_by', b'eloqua')]]
        produce_message(*test_input)
        assert producer._value_serializer == producer_serializer
        metrics_manager.inc_messages_produced.assert_called_with(1, 'test_topic_1')
        producer.produce.assert_called_with(
            **dict(
                topic='test_topic_1',
                key='blah_key',
                value={'blah_value': 'blah_value'},
                headers={'guid': '112233', 'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}))

    def test_produce_message__dict_header_passthrough(self, producer, producer_serializer):
        """
        Happy path, check that header passthrough gets the header kw arg updates
        """
        test_input = [
            producer,
            dict(topic='test_topic_1',
                 key='blah_key',
                 value={'blah_value': 'blah_value'},
                 headers={'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}),
            None,
            {'guid': '112233', 'last_updated_by': 'eloqua'}]
        produce_message(*test_input)
        assert producer._value_serializer == producer_serializer
        producer.produce.assert_called_with(
            **dict(
                topic='test_topic_1',
                key='blah_key',
                value={'blah_value': 'blah_value'},
                headers={'guid': '112233', 'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}))

    def test_produce_message__has_topic_attribute(self, producer, producer_serializer):
        producer.topic = 'test_topic_1'
        test_input = [
            producer,
            dict(key='blah_key',
                 value={'blah_value': 'blah_value'},
                 headers={'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}),
            None,
            [('guid', b'112233'), ('last_updated_by', b'eloqua')]]
        produce_message(*test_input)
        assert producer._value_serializer == producer_serializer
        producer.produce.assert_called_with(
            **dict(
                topic='test_topic_1',
                key='blah_key',
                value={'blah_value': 'blah_value'},
                headers={'guid': '112233', 'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}))

    def test_produce_message__bad_headers(self, producer, producer_serializer):
        test_input = [
            producer,
            dict(
                key='blah_key',
                value={'blah_value': 'blah_value'},
                header={'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}),
            None,
            [('guid', b'112233'), ('last_updated_by', b'eloqua')]]
        with pytest.raises(KeyError):
            produce_message(*test_input)

        test_input = [
            producer,
            dict(
                key='blah_key',
                value={'blah_value': 'blah_value'},
                headers={'last_updated_by': 'sfdc', 'new_header_key': 'new_header_value'}),
            None,
            [('last_updated_by', b'eloqua')]]
        with pytest.raises(ProduceHeadersException):
            produce_message(*test_input)
            
    def test_confirm_produce(self, producer):
        producer.__len__.side_effect = [2, 1, 0]
        confirm_produce(producers=producer)
        producer.flush.assert_has_calls([call(timeout=20)]*2)

        producer.__len__.side_effect = [4, 3, 2, 1, 0]
        with pytest.raises(ProducerTimeoutFailure):
            confirm_produce(producers=producer)

    def test_confirm_produce__multi(self, producer):
        producer2 = deepcopy(producer)
        producer.__len__.side_effect = [2, 1, 0, 2, 1, 0]
        confirm_produce(producers={'test_topic_1': producer, 'test_topic_2': producer2})
        producer.flush.assert_has_calls([call(timeout=20)]*2)
        producer2.flush.assert_has_calls([call(timeout=20)]*2)


if __name__ == '__main__':
    pytest.main([__file__])



import pytest
from copy import deepcopy
from unittest import TestCase
from unittest.mock import patch
from nubium_utils.env_var_generator import EnvVars


class TestEnvVarGenerator(TestCase):

    def setUp(self):
        self.dict_out = {"test_env_field": "test_env_value"}
        self.env_vars = EnvVars(self.dict_func)

    def dict_func(self):
        return self.dict_out

    def test_it_works(self):
        self.assertDictEqual(self.env_vars(), self.dict_out)

    def test_returns_same_python_object_each_call(self):
        self.assertEqual(id(self.env_vars()), id(self.env_vars()))  # also ensures subsequent calls work!

    def test__env_reload(self):
        self.env_vars()
        self.dict_out.update({'I_SHOULD_BE_ADDED_AFTER_RELOAD': 'plz'})
        assert 'I_SHOULD_BE_ADDED_AFTER_RELOAD' not in self.env_vars()
        self.env_vars._reload()
        assert 'I_SHOULD_BE_ADDED_AFTER_RELOAD' in self.env_vars()

if __name__ == '__main__':
    pytest.main([__file__])


import setuptools

with open("README.md", "r") as file_obj:
    long_description = file_obj.read()

install_requires = [
    "confluent-kafka[avro]==1.7.0",
    "lbry-rocksdb",
    "nubium-schemas",
    "prometheus_client",
    "psutil",
    "python-box",
    "python-dateutil",
    "python-dotenv",
    "python_eloqua_wrapper",
    "pytz",
    "pyyaml",
    "requests",
    "virtualenv",
    "virtualenv-api",
]

dev_requires = install_requires + [
    "pip-tools",
    "pytest",
    "pytest-cov",
    "twine",
    "time-machine",
]

packages = setuptools.find_packages()

setuptools.setup(
    name="nubium-utils",
    version="2.2.0",
    author="Edward Brennan",
    author_email="ebrennan@redhat.com",
    description="Some Kafka utility functions and patterns for the nubium project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.corp.redhat.com/mkt-ops-de/nubium-utils.git",
    packages=packages,
    install_requires=install_requires,
    include_package_data=True,
    extras_require={"dev": dev_requires},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from copy import deepcopy

class EnvVars:
    def __init__(self, dict_func):
        self._dict_func = dict_func
        self._gen = None
        self._env_dict = lambda: None
        self._reload()
    
    def __repr__(self):
        return str(self._env_dict())
        
    def __call__(self, *args, **kwargs):
        return self._env_dict()
    
    def _get_generator(self):
        d = deepcopy(self._dict_func())
        while True:
            yield d

    def _reload(self):  # for easy unit testing only
        gen = self._get_generator()
        self._env_dict = lambda: next(gen)
